import datetime
from typing import Any, Callable, Dict, Optional, Tuple, Union
from uuid import uuid4

from crontab import CronTab
from loguru import logger
from tortoise import timezone

from rearq.job import JobStatus
from rearq.server.models import Job, JobResult
from rearq.utils import ms_to_datetime, timestamp_ms_now, to_ms_timestamp


class Task:
    def __init__(
        self,
        bind: bool,
        function: Callable,
        queue: str,
        rearq,
        job_retry: int,
        job_retry_after: int,
        expire: Optional[Union[float, datetime.datetime]] = None,
    ):

        self.job_retry = job_retry
        self.job_retry_after = job_retry_after
        self.queue = queue
        self.rearq = rearq
        self.function = function
        self.bind = bind
        self.expire = expire

    async def delay(
        self,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        job_id: str = None,
        countdown: Union[float, datetime.timedelta] = 0,
        eta: Optional[datetime.datetime] = None,
        expire: Optional[Union[float, datetime.datetime]] = None,
        job_retry: int = 0,
        job_retry_after: int = 60,
    ) -> Job:
        """
        Add job to queue.
        :param args: Job args.
        :param kwargs: Job kwargs.
        :param job_id: Custom job id.
        :param countdown: Delay seconds to execute.
        :param eta: Delay to datetime to execute.
        :param expire: Override default expire.
        :param job_retry: Override default job retry.
        :param job_retry_after: Override default job retry after.
        :return:
        """
        if not job_id:
            job_id = uuid4().hex
        expire_time = None
        expires = expire or self.expire
        if expires:
            expire_time = ms_to_datetime(to_ms_timestamp(expires))

        job = await Job.get_or_none(job_id=job_id)
        if job:
            logger.warning(f"Job {job_id} exists")
            return job

        job = Job(
            task=self.function.__name__,
            args=args,
            kwargs=kwargs,
            job_retry=job_retry or self.job_retry,
            queue=self.queue,
            job_id=job_id,
            expire_time=expire_time,
            enqueue_time=timezone.now(),
            job_retry_after=job_retry_after,
        )

        if not eta and not countdown:
            job.status = JobStatus.queued
            await job.save()
            await self.rearq.redis.xadd(self.queue, {"job_id": job_id})
        else:
            if countdown:
                defer_ms = to_ms_timestamp(countdown)
            elif eta:
                defer_ms = to_ms_timestamp(eta)
            else:
                defer_ms = timestamp_ms_now()
            job.status = JobStatus.deferred
            await job.save()
            await self.rearq.zadd(defer_ms, f"{self.queue}:{job_id}")
            await self.rearq.pub_delay(defer_ms)
        return job


async def check_pending_msgs(self: Task, queue: str, group_name: str, timeout: int):
    """
    check pending messages
    :return:
    """
    redis = self.rearq.redis
    pending = await redis.xpending(self.queue, group_name)
    count = pending.get("pending")
    if not count:
        return
    pending_msgs = await redis.xpending_range(
        self.queue,
        group_name,
        min=pending.get("min"),
        max=pending.get("max"),
        count=count,
    )
    p = redis.pipeline()
    execute = False
    for msg in pending_msgs:
        msg_id = msg.get("message_id")
        idle_time = msg.get("time_since_delivered")
        if int(idle_time / 10 ** 6) > timeout * 2:
            execute = True
            p.xack(queue, group_name, msg_id)
            job_result = await JobResult.filter(msg_id=msg_id).only("job_id").first()
            if job_result:
                p.xadd(queue, {"job_id": job_result.job_id})
    if execute:
        return await p.execute()


async def check_keep_job(self: Task):
    rearq = self.rearq
    keep_job_days = rearq.keep_job_days
    time = timezone.now() - datetime.timedelta(days=keep_job_days)
    return await Job.filter(
        status__in=[JobStatus.failed, JobStatus.success, JobStatus.expired], enqueue_time__lt=time
    ).delete()


class CronTask(Task):
    _cron_tasks: Dict[str, "CronTask"] = {}
    next_run: int

    def __init__(
        self,
        bind: bool,
        function: Callable,
        queue: str,
        rearq,
        job_retry: int,
        job_retry_after: int,
        cron: str,
        expire: Optional[Union[float, datetime.datetime]] = None,
        run_at_start: Optional[bool] = False,
    ):
        super().__init__(bind, function, queue, rearq, job_retry, job_retry_after, expire)
        self.crontab = CronTab(cron)
        self.cron = cron
        self.run_at_start = run_at_start
        self.set_next()

    def set_next(self):
        self.next_run = to_ms_timestamp(self.crontab.next(default_utc=False))

    @classmethod
    def add_cron_task(cls, function: str, cron_task: "CronTask"):
        cls._cron_tasks[function] = cron_task

    @classmethod
    def get_cron_tasks(cls):
        return cls._cron_tasks


def is_built_task(task: str):
    return task in [check_pending_msgs.__name__, check_keep_job.__name__]



import asyncio
import datetime
import hashlib
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Set, Union

import aioredis
import aioredis.sentinel
from aioredis import Redis

from rearq import constants
from rearq.constants import DELAY_QUEUE_CHANNEL
from rearq.exceptions import UsageError
from rearq.task import CronTask, Task, is_built_task

Serializer = Callable[[Dict[str, Any]], bytes]
Deserializer = Callable[[bytes], Dict[str, Any]]


class ReArq:
    _redis: Optional[Redis] = None
    _task_map: Dict[str, Task] = {}
    _queue_task_map: Dict[str, List[str]] = {}
    _on_startup: Set[Callable] = set()
    _on_shutdown: Set[Callable] = set()

    def __init__(
        self,
        redis_url: str = "redis://localhost:6379/0",
        sentinels: Optional[List[str]] = None,
        sentinel_master: str = "master",
        job_retry: int = 3,
        job_retry_after: int = 60,
        max_jobs: int = 10,
        job_timeout: int = 300,
        expire: Optional[Union[float, datetime.datetime]] = None,
        delay_queue_num: int = 1,
        keep_job_days: Optional[int] = None,
    ):
        """
        :param sentinel_master:
        :param job_retry: Default job retry times.
        :param job_retry_after: Default job retry after seconds.
        :param max_jobs: Max concurrent jobs.
        :param job_timeout: Job max timeout.
        :param expire: Job default expire time.
        :param delay_queue_num: How many key to store delay tasks, for large number of tasks, split it to improve performance.
        """
        self.job_timeout = job_timeout
        self.max_jobs = max_jobs
        self.job_retry = job_retry
        self.job_retry_after = job_retry_after
        self.expire = expire
        self.sentinels = sentinels
        self.sentinel_master = sentinel_master
        self.redis_url = redis_url
        self.delay_queue_num = delay_queue_num
        self.keep_job_days = keep_job_days

    async def init(self):
        if self._redis:
            return

        if self.sentinels:
            sentinel = aioredis.sentinel.Sentinel(self.sentinels, decode_responses=True)
            redis = sentinel.master_for(self.sentinel_master)

        else:
            redis = aioredis.from_url(
                self.redis_url,
                decode_responses=True,
            )
        self._redis = redis

    @property
    def redis(self):
        if not self._redis:
            raise UsageError("You must call .init() first!")
        return self._redis

    @property
    def task_map(self) -> Dict[str, Task]:
        return self._task_map

    @property
    def queue_task_map(self) -> Dict[str, List[str]]:
        return self._queue_task_map

    def get_queue_tasks(self, queue: str) -> List[str]:
        tasks = self._queue_task_map.get(queue)
        return tasks or []

    def create_task(
        self,
        bind: bool,
        func: Callable,
        queue: Optional[str] = None,
        cron: Optional[str] = None,
        name: Optional[str] = None,
        job_retry: Optional[int] = None,
        job_retry_after: Optional[int] = None,
        expire: Optional[Union[float, datetime.datetime]] = None,
        run_at_start: Optional[bool] = False,
    ):

        if not callable(func):
            raise UsageError("Task must be Callable!")

        function = name or func.__name__
        if function in (self._queue_task_map.get(queue) or []):
            raise UsageError("Task name must be unique!")

        if not is_built_task(function):
            self._queue_task_map.setdefault(queue, []).append(function)

        defaults = dict(
            function=func,
            queue=constants.QUEUE_KEY_PREFIX + queue if queue else constants.DEFAULT_QUEUE,
            rearq=self,
            job_retry=job_retry or self.job_retry,
            job_retry_after=job_retry_after or self.job_retry_after,
            expire=expire or self.expire,
            bind=bind,
        )
        if cron:
            t = CronTask(**defaults, cron=cron, run_at_start=run_at_start)
            CronTask.add_cron_task(function, t)
        else:
            t = Task(**defaults)
        self._task_map[function] = t
        return t

    def task(
        self,
        bind: bool = True,
        queue: Optional[str] = None,
        cron: Optional[str] = None,
        name: Optional[str] = None,
        job_retry: Optional[int] = None,
        job_retry_after: Optional[int] = None,
        expire: Optional[Union[float, datetime.datetime]] = None,
        run_at_start: Optional[bool] = False,
    ):
        """
        Task decorator
        :param bind: Bind task obj to first argument.
        :param queue: Default queue.
        :param cron: See https://github.com/josiahcarlson/parse-crontab, if defined, which is a cron task.
        :param name: Task name, default is function name.
        :param job_retry: Override default job retry.
        :param job_retry_after: Override default job retry after.
        :param expire: Override default expire.
        :param run_at_start: Whether run at startup, only available for cron task.
        :return:
        """
        if not cron and run_at_start:
            raise UsageError("run_at_start only work in cron task")

        def wrapper(func: Callable):
            return self.create_task(
                bind, func, queue, cron, name, job_retry, job_retry_after, expire, run_at_start
            )

        return wrapper

    def on_startup(self, func: Callable):
        self._on_startup.add(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def on_shutdown(self, func: Callable):
        self._on_shutdown.add(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    async def shutdown(self):
        tasks = []
        for fun in self._on_shutdown:
            tasks.append(fun())
        if tasks:
            await asyncio.gather(*tasks)

    async def startup(self):
        tasks = []
        for fun in self._on_startup:
            tasks.append(fun())
        if tasks:
            await asyncio.gather(*tasks)

    async def close(self):
        if not self._redis:
            return
        await self._redis.close()
        self._redis = None

    def get_delay_queue(self, data: str):
        num = int(hashlib.md5(data.encode()).hexdigest(), 16) % self.delay_queue_num  # nosec:B303
        return f"{constants.DELAY_QUEUE}:{num}"

    @property
    def delay_queues(self):
        for num in range(self.delay_queue_num):
            yield f"{constants.DELAY_QUEUE}:{num}"

    async def zadd(self, score: int, data: str):
        queue = self.get_delay_queue(data)
        return await self.redis.zadd(queue, mapping={data: score})

    async def pub_delay(self, ms: float):
        return await self.redis.publish(DELAY_QUEUE_CHANNEL, ms)
